@@ -448,7 +448,7 @@ local function pick_call_hierarchy_item(call_hierarchy_items)
   if choice < 1 or choice > #items then
     return
   end
-  return choice
+  return call_hierarchy_items[choice]
 end
 
 --- @param method string
@@ -836,14 +836,10 @@ function M.code_action(opts)
   if opts.diagnostics or opts.only then
     opts = { options = opts }
   end
-  local context = opts.context or {}
+  local context = opts.context and vim.deepcopy(opts.context) or {}
   if not context.triggerKind then
     context.triggerKind = vim.lsp.protocol.CodeActionTriggerKind.Invoked
   end
-  if not context.diagnostics then
-    local bufnr = api.nvim_get_current_buf()
-    context.diagnostics = vim.lsp.diagnostic.get_line_diagnostics(bufnr)
-  end
   local mode = api.nvim_get_mode().mode
   local bufnr = api.nvim_get_current_buf()
   local win = api.nvim_get_current_win()
@@ -885,7 +881,23 @@ function M.code_action(opts)
     else
       params = util.make_range_params(win, client.offset_encoding)
     end
-    params.context = context
+    if context.diagnostics then
+      params.context = context
+    else
+      local ns_push = vim.lsp.diagnostic.get_namespace(client.id, false)
+      local ns_pull = vim.lsp.diagnostic.get_namespace(client.id, true)
+      local diagnostics = {}
+      local lnum = api.nvim_win_get_cursor(0)[1] - 1
+      vim.list_extend(diagnostics, vim.diagnostic.get(bufnr, { namespace = ns_pull, lnum = lnum }))
+      vim.list_extend(diagnostics, vim.diagnostic.get(bufnr, { namespace = ns_push, lnum = lnum }))
+      params.context = vim.tbl_extend('force', context, {
+        ---@diagnostic disable-next-line: no-unknown
+        diagnostics = vim.tbl_map(function(d)
+          return d.user_data.lsp
+        end, diagnostics),
+      })
+    end
+
     client.request(ms.textDocument_codeAction, params, on_result, bufnr)
   end
 end
